import pandas as pd
from datetime import datetime
import logging
import subprocess


def process_data(data):
    record = []
    record_dict = {}
    author_dict = {}

    for package_name, info in data.items():
        repo_name = info.get("repo_name", "")
        pkg_category = info.get("category", "")
        old_version = info.get("tag1", "")
        new_version = info.get("tag2", "")
        repo_link = info.get("repo_link", "")

        # The 'or' is to handle the case where the information returned is None
        commits = info.get("authors", []) or []

        if not commits:
            record.append(
                {
                    "package_name": f"`{package_name}`",
                    "repo_name": repo_name,
                    "repo_link": repo_link,
                    "category": pkg_category,
                    "old_version": old_version,
                    "new_version": new_version,
                    "sha": None,
                    "author": None,
                    "author_first": None,
                    "merger": None,
                    "prr_first": None,
                    "reviewer": None,
                    "reviewer_type": None,
                }
            )

        for commit in commits:
            sha = commit.get("sha", "")
            author = commit.get("login", "")
            author_first = commit.get("commit_result", {}).get("is_first_commit", "")

            commit_merge_infos = commit.get("commit_merged_info", [])

            prr_first = None
            reviewer = None
            reviewer_type = None
            merger = None

            for commit_merge_info in commit_merge_infos:
                merger = commit_merge_info.get("merge_by", "")
                if "reviews" in commit_merge_info and isinstance(commit_merge_info["reviews"], list):
                    for review in commit_merge_info["reviews"]:
                        if isinstance(review, dict) and "prr_data" in review:
                            reviewer = review.get("review_author")
                            reviewer_type = review.get("review_author_type")
                            prr_data = review.get("prr_data", {})
                            if prr_data:
                                prr_first = prr_data.get("is_first_prr")

                if sha in record_dict:
                    if package_name not in record_dict[sha]["package_name"]:
                        record_dict[sha]["package_name"].append(package_name)
                else:
                    record_dict[sha] = {
                        "package_name": [package_name],
                        "repo_name": repo_name,
                        "old_version": old_version,
                        "new_version": new_version,
                        "sha": sha,
                        "author": author,
                        "author_first": author_first,
                        "merger": merger,
                        "prr_first": prr_first,
                        "reviewer": reviewer,
                        "reviewer_type": reviewer_type,
                    }

                package_number = len(record_dict[sha]["package_name"])
                record_dict[sha]["package_number"] = package_number

            record.append(
                {
                    "package_name": package_name,
                    "repo_name": repo_name,
                    "category": pkg_category,
                    "old_version": old_version,
                    "new_version": new_version,
                    "sha": sha,
                    "author": author,
                    "author_first": author_first,
                    "merger": merger,
                    "prr_first": prr_first,
                    "reviewer": reviewer,
                    "reviewer_type": reviewer_type,
                }
            )

    record_list = list(record_dict.values())
    author_list = list(author_dict.values())

    return record, record_list, author_list


def create_dataframe(
    record,
    default_columns=[
        "package_name",
        "repo_name",
        "repo_link",
        "category",
        "old_version",
        "new_version",
        "sha",
        "author",
        "author_first",
        "merger",
        "prr_first",
        "reviewer",
        "reviewer_type",
        "signature_changes",
    ],
):
    df = pd.DataFrame(record)
    # if empty, have the columns in the df regardless
    for col in default_columns:
        if col not in df.columns:
            df[col] = None
    return df


def filter_df(df):
    df_author_first = df[(df["author_first"] == True)]
    df_review_first = df[df["prr_first"] == True]
    df_both_first = df[(df["author_first"] == True) & (df["prr_first"] == True)]

    return df_author_first, df_review_first, df_both_first


def print_check_info(df, summary, md_file, amount):
    if amount > 0:
        md_file.write("\n")
        md_file.write(
            f"""
<details>
    <summary>{summary} ({amount})</summary>
        """
        )
        md_file.write("\n\n\n")
        md_file.write(df.to_markdown(index=False))
        md_file.write("\n</details>\n")
        return True
    return False


def generate_diff_report(
    data, project_repo_name, release_version_old, release_version_new, gradual_report, output_file
):
    logging.info(f"Generating differential report for {project_repo_name}")
    record, record_list, author_list = process_data(data)

    df_all = create_dataframe(record)

    # To know downgraded packages
    downgraded = df_all["category"] == "Downgraded package"
    downgraded_number = df_all[downgraded]["package_name"].nunique()
    selected_columns = [
        "package_name",
        "repo_link",
        "category",
        "old_version",
        "new_version",
    ]
    df_down_selected = df_all[downgraded][selected_columns]
    df_down_selected_without_index = df_down_selected.reset_index(drop=True)

    # We want a table that indexes the author
    df_author = create_dataframe(record_list).set_index("author")

    new_order = ["sha"] + [col for col in df_author.columns if col != "sha"]
    df_author = df_author.reindex(columns=new_order)

    df_author_new_author, df_author_new_reviewer, df_author_both_new = filter_df(df_author)

    # know the columns

    profile_url = "https://github.com/"

    cp_df_author_new_author = df_author_new_author.copy()
    cp_df_author_new_author.index = df_author_new_author.index.map(lambda x: f"[{x}]({profile_url}{x})")
    # df_author_new_author.loc[:,"author"] = df_author_new_author.loc[:,"author"].map(lambda x: f"[{x}]({profile_url}{x})")

    cp_df_author_new_reviewer = df_author_new_reviewer.copy()
    cp_df_author_new_reviewer["reviewer"] = df_author_new_reviewer["reviewer"].apply(
        lambda x: f"[{x}]({profile_url}{x})"
    )

    cp_df_author_both_new = df_author_both_new.copy()
    cp_df_author_both_new.index = df_author_both_new.index.map(lambda x: f"[{x}]({profile_url}{x})")
    cp_df_author_both_new["reviewer"] = df_author_both_new["reviewer"].apply(lambda x: f"[{x}]({profile_url}{x})")

    new_author_commits = df_author_new_author.shape[0]
    new_reviewer_commits = df_author_new_reviewer.shape[0]
    both_new_commits = df_author_both_new.shape[0]

    signature_changes = df_all[df_all["category"].str.contains("signature changes", case=False)]
    signature_changes_df = signature_changes[["package_name", "old_version", "new_version", "signature_changes"]]
    signature_changes_number = signature_changes["package_name"].nunique()

    common_counts = {
        "### Total packages in the supply chain": len(data),
    }

    reports = {
        "signature_changes": {
            "amount": signature_changes_number,
            "df": signature_changes_df,
            "summary": ":lock: Packages with signature changes (⚠️⚠️⚠️)",
        },
        "downgraded": {
            "amount": downgraded_number,
            "df": df_down_selected_without_index,
            "summary": ":heavy_exclamation_mark: Downgraded packages (⚠️⚠️)",
        },
        "both_new": {
            "amount": both_new_commits,
            "df": cp_df_author_both_new,
            "summary": ":alien: Commits made by both New Authors and Reviewers (⚠️⚠️)",
        },
        "new_reviewer": {
            "amount": new_reviewer_commits,
            "df": cp_df_author_new_reviewer,
            "summary": ":see_no_evil: Commits approved by New Reviewers (⚠️⚠️)",
        },
        "new_author": {
            "amount": new_author_commits,
            "df": cp_df_author_new_author,
            "summary": ":neutral_face: Commits made by New Authors (⚠️)",
        },
    }

    # We write into a markdown file
    with open(output_file, "w", encoding="utf-8") as f:
        preamble = f"""
# Software Supply Chain Report of {project_repo_name} - {release_version_old} &rarr; {release_version_new}
"""
        if gradual_report:
            preamble += """
\nThis report is a gradual report: that is, only the highest severity smell type with issues found within this project is reported.
Gradual reports are enabled by default. You can disable this feature, and get a full report, by using the `--gradual-report=false` flag.
"""
        preamble += "\n"
        f.write(preamble)

        f.write(
            """
<details>
    <summary>How to read the results :book: </summary>
    \n Dirty-waters has analyzed your project dependencies and found different categories for each of them:\n
    \n - ⚠️⚠️⚠️ : high severity \n
    \n - ⚠️⚠️: medium severity \n
    \n - ⚠️: low severity \n
</details>
        """
        )

        f.write("\n")

        for key, value in common_counts.items():
            f.write(f"\n {key}: {value}\n")

        for info in reports.values():
            f.write(f"\n {info['summary']}: {info['amount']}\n")

        f.write("\n")
        f.write("### Fine grained information\n")
        f.write(
            "\n:dolphin: For further information about software supply chain smells in your project, take a look at the following tables.\n"
        )

        for info in reports.values():
            printed = print_check_info(info["df"], info["summary"], f, info["amount"])
            if gradual_report and printed:
                f.write("\n")
                break

        f.write("\n### Call to Action:\n")
        f.write(
            """
                      
<details>
<summary>👻What do I do now? </summary>

For packages **with signature changes**:\n
- **Why?** Changes in code signatures could indicate tampering with the package or compromised build processes, potentially introducing malicious code.\n
1. This means that a dependency either had code signature and now does not, or that the signature was valid and now it's not.
2. This could be a security risk, and you should halt the project until you can verify the changes. \n

For **downgraded dependencies**:\n
- **Why?** Downgrading packages may reintroduce known security vulnerabilities that were fixed in newer versions.\n
1. Check the release notes of the new version to see if the downgrade is intentional. If the new version is more than one release ahead, verify whether any breaking changes in between apply to your project.
2. If the downgrade is unintentional, consider updating the package to a version that is compatible with your project.

For commits made by **both new authors and reviewers**:\n
- **Why?** When both authors and reviewers are new to a project, there's a higher risk of malicious code being introduced due to lack of established trust and project knowledge.\n
1. Verify, as best as you can, that the new authors and reviewers are not malicious actors.
2. If you are unsure, consider reverting the changes.

For commits approved by **new reviewers**:\n
- **Why?** New reviewers may not be familiar with the project's security requirements or may not have the expertise to identify malicious code.\n
1. Verify, as best as you can, that the new reviewers are not malicious actors.

For commits made by **new authors**:\n
- **Why?** New contributors could potentially introduce security vulnerabilities, either accidentally or intentionally.\n
1. Verify, as best as you can, that the new authors are not malicious actors.
2. The fact that the reviewers are not new to the repository is a good sign.
</details>
"""
        )

        f.write(f"---\n")
        f.write(f"\nReport created by [dirty-waters](https://github.com/chains-project/dirty-waters/).\n")
        f.write(f"\nReport created on {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n")

        # Tool version
        tool_commit_hash = subprocess.check_output(["git", "rev-parse", "--short", "HEAD"]).strip().decode("utf-8")
        f.write(f"- Tool version: {tool_commit_hash}\n")
        f.write(f"- Project Name: {project_repo_name}\n")
        f.write(f"- Compared project versions: {release_version_old} & {release_version_new}\n")
    print(f"Report from differential analysis generated at {output_file}")



import requests
import sqlite3
import os
from pathlib import Path
import json
import copy
import logging
from tool.tool_config import get_cache_manager, make_github_request

cache_manager = get_cache_manager()

GITHUB_TOKEN = os.getenv("GITHUB_API_TOKEN")

headers = {
    "Authorization": f"Bearer {GITHUB_TOKEN}",
    "Accept": "application/vnd.github.v4+json",
}

url = "https://api.github.com/graphql"


def get_multiple_pr_info(repo_name, review_author_logins):
    # Build dynamic query with aliases
    query_fragments = []
    variables = {}

    for i, login in enumerate(review_author_logins):
        alias = f"search_{i}"
        query_fragments.append(
            f"""
        {alias}: search(query: $query_{i}, type: ISSUE, last: 1) {{
            nodes {{
                ... on PullRequest {{
                    mergedAt
                    merged
                    mergedBy {{
                        login
                    }}
                    authorAssociation
                    reviews(first:1, states:APPROVED) {{
                        edges {{
                            node {{
                                id
                                author {{
                                    login
                                    __typename
                                    url
                                }}
                                authorAssociation
                                createdAt
                                publishedAt
                                submittedAt
                                state
                                repository {{
                                    owner {{
                                        login
                                    }}
                                    name
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        )
        variables[f"query_{i}"] = f"repo:{repo_name} is:pr reviewed-by:{login} sort:author-date-asc"

    # Combine all query fragments
    complete_query = """
    query({}) {{
        {}
    }}
    """.format(
        ", ".join(f"$query_{i}: String!" for i in range(len(review_author_logins))), "\n".join(query_fragments)
    )

    body = {"query": complete_query, "variables": variables}
    response = make_github_request(url, method="POST", json_data=body, headers=headers)

    # Restructure response to match expected format
    if "data" in response:
        queries = []
        for i in range(len(review_author_logins)):
            search_data = response["data"].get(f"search_{i}")
            if search_data:
                queries.append(search_data)
        return {"data": {"queries": queries}}
    return response


def get_pr_review_info(data):
    logging.info("Getting PR review info...")
    pr_data = copy.deepcopy(data)

    # Collect all uncached reviewer lookups needed
    uncached_lookups = []
    for package, info in pr_data.items():
        for author in info.get("authors", []):
            for merge_info in author.get("commit_merged_info", []):
                if merge_info.get("state") != "MERGED":
                    continue

                repo_name = merge_info.get("repo")
                for reviewer in merge_info.get("reviews", []):
                    review_author_login = reviewer.get("review_author")
                    if not review_author_login:
                        continue

                    if not cache_manager.github_cache.get_pr_review(repo_name, review_author_login):
                        uncached_lookups.append((repo_name, review_author_login))

    # Batch fetch uncached reviewers by repository
    by_repo = {}
    for repo_name, login in uncached_lookups:
        by_repo.setdefault(repo_name, set()).add(login)

    for repo_name, logins in by_repo.items():
        response = get_multiple_pr_info(repo_name, list(logins))
        # Cache individual results
        for login, result in zip(logins, response.get("data", {}).get("queries", [])):
            cache_manager.github_cache.cache_pr_review(package, repo_name, login, {"data": {"search": result}})

    # Process the data using cached results
    for package, info in pr_data.items():
        authors = info.get("authors", [])
        if authors:
            for author in authors:
                merge_infos = author.get("commit_merged_info", [])
                merge_info = merge_infos[0]
                repo_name = merge_info.get("repo")
                commit_sha = merge_info.get("commit_sha")
                merge_state = merge_info.get("state")
                reviewer_info = merge_info.get("reviews", [])

                review_author_login = "no_reviewer"
                review_id = None
                first_pr_info = None

                if merge_state == "MERGED" and len(reviewer_info) >= 1:
                    for reviewer in reviewer_info:
                        review_author_login = reviewer.get("review_author")
                        review_id = reviewer.get("review_id")
                        first_pr_info = cache_manager.github_cache.get_pr_review(repo_name, review_author_login)
                        useful_info = first_pr_info.get("data", {}).get("search", {}).get("nodes", [])
                        first_review_info = useful_info[0] if useful_info else {}
                        all_useful_first_prr_info = first_review_info.get("reviews", {}).get("edges", [])

                        if len(all_useful_first_prr_info) >= 1:
                            first_review = (
                                all_useful_first_prr_info[0].get("node", {}) if all_useful_first_prr_info else {}
                            )
                            first_prr_node_id = first_review.get("id")
                            first_prr_author = first_review.get("author", {})
                            first_prr_author = first_prr_author.get("login") if first_prr_author else None
                            first_prr_state = first_review.get("state")
                            first_prr_author_association = first_review.get("authorAssociation")
                            is_first_prr = False

                            if review_id is not None:
                                if first_prr_node_id == review_id:
                                    is_first_prr = True
                            else:
                                is_first_prr = "No review info"

                            useful_pr_info = {
                                "package": package,
                                "repo": repo_name,
                                "author_from_review": review_author_login,
                                "commit_sha": commit_sha,
                                "merge_state": merge_state,
                                "review_id": review_id,
                                "first_prr_node_id": first_prr_node_id,
                                "first_prr_author": first_prr_author,
                                "first_prr_state": first_prr_state,
                                "first_prr_repo": repo_name,
                                "first_prr_review_author_association": first_prr_author_association,
                                "is_first_prr": is_first_prr,
                            }
                        else:
                            useful_pr_info = None

                        reviewer["prr_data"] = useful_pr_info

        else:
            logging.info(f"No authors for package:{package}")
            info["prr_data"] = None

    logging.info("PR review info processed.")
    return pr_data



import json
import os
import base64
import time
import urllib.parse
from tqdm import tqdm
import git

import requests
import subprocess
import re

from tool.tool_config import get_cache_manager, make_github_request
from tool.compare_commits import tag_format as construct_tag_format, find_existing_tags_batch
import logging
import xmltodict

github_token = os.getenv("GITHUB_API_TOKEN")

headers = {
    "Authorization": f"Bearer {github_token}",
    "Accept": "application/vnd.github.v3+json",
}

cache_manager = get_cache_manager()


MAX_WAIT_TIME = 15 * 60

DEFAULT_ENABLED_CHECKS = {
    "source_code": True,
    "source_code_sha": True,
    "deprecated": True,
    "forks": False,
    "provenance": True,
    "code_signature": True,
    "aliased_packages": True,
}

SCHEMAS_FOR_CACHE_ANALYSIS = {
    "source_code": {
        "is_github": False,
        "github_api": "",
        "github_url": "",
        "github_exists": None,
        "github_redirected": None,
        "redirected_repo": "",
        "status_code": 200,
        "archived": None,
        "is_fork": None,
        "source_code_version": {
            "exists": None,
            "tag_version": "",
            "is_sha": None,
            "sha": "",
            "url": "",
            "message": "",
            "status_code": 404,
        },
        "parent_repo_link": "",
        "open_issues_count": 0,
        "error": "No error message.",
    },
    "package_info": {
        "package_only_name": "",
        "package_version": "",
        "deprecated_in_version": None,
        "provenance_in_version": None,
        "all_deprecated": None,
        "provenance_url": None,
        "provenance_info": None,
        "status_code": 200,
    },
    "code_signature": {
        "signature_present": None,
        "signature_valid": None,
    },
}


def update_package_info(package_info, field, new_data):
    # Also updates in place
    if field not in package_info:
        package_info[field] = {}
    package_info[field].update(new_data)
    return package_info


def check_deprecated_and_provenance(package, package_version, pm):
    """
    Check if the package is deprecated and if it has a provenance from the package manager's registry.
    """

    def check_npm(package, package_version):
        try:
            response = requests.get(f"https://registry.npmjs.org/{package}", timeout=20)

            response.raise_for_status()
        except requests.RequestException:
            return {
                "package_only_name": package,
                "package_version": package_version,
                "error": "Failed to fetch package data",
                "status_code": response.status_code if response else "No Response",
            }

        data = response.json()

        if_deprecated = False
        has_provenance = False
        provenance_url = None
        provenance_info = None
        all_deprecated = True

        version_info = data.get("versions", {}).get(package_version, {})
        all_versions = data.get("versions", {})

        deprecated_in_version = version_info.get("deprecated", "")
        provenance_in_version = version_info.get("dist", {}).get("attestations", "")

        if deprecated_in_version:
            if_deprecated = True

        if provenance_in_version:
            has_provenance = True
            provenance_url = provenance_in_version.get("url")
            provenance_info = provenance_in_version.get("provenance")

        for version in all_versions.values():
            if not version.get("deprecated"):
                all_deprecated = False
                break

        npm_package_info = {
            "package_only_name": package,
            "package_version": package_version,
            "deprecated_in_version": if_deprecated,
            "provenance_in_version": has_provenance,
            "all_deprecated": all_deprecated,
            "provenance_url": provenance_url,
            "provenance_info": provenance_info,
            "status_code": 200,
        }

        return npm_package_info

    def check_maven(package, package_version):
        return {
            "package_only_name": package,
            "package_version": package_version,
            "status_code": 404,
            "error": "Maven does not have a registry",
        }

    if pm in ("yarn-berry", "yarn-classic", "pnpm", "npm"):
        return check_npm(package, package_version)
    elif pm == "maven":
        # maven doesn't have this
        return check_maven(package, package_version)
    else:
        # log stuff
        # blow up
        logging.error(f"Package manager {pm} not supported.")


def check_code_signature(package_name, package_version, pm):
    def check_maven_signature(package_name, package_version):
        # Construct the command
        command = f"mvn org.simplify4u.plugins:pgpverify-maven-plugin:1.18.2:show -Dartifact={package_name}:{package_version}"

        # Run the command
        output = subprocess.run(command, shell=True, capture_output=True, text=True)
        ansi_escape = re.compile(r"\x1b\[[0-9;]*[a-zA-Z]")
        escaped_output = ansi_escape.sub("", output.stdout)

        # Regular expression to extract the PGP signature section
        pgp_signature_pattern = re.compile(r"PGP signature:\n(?:[ \t]*.+\n)*?[ \t]*status:\s*(\w+)", re.MULTILINE)
        match = pgp_signature_pattern.search(escaped_output)
        logging.info(f"Code Signature match: {match}")
        if match:
            # Extract the status
            status = match.group(1).strip().lower()
            return {"signature_present": True, "signature_valid": status == "valid"}

        # If no match is found, return no PGP signature present
        logging.warning(f"No match. Command output was\n{escaped_output!r}")
        return {"signature_present": False, "signature_valid": False}

    def check_npm_signature(package, package_version):
        # NOTE: for future reference, NPM migrated from PGP signatures to ECDSA registry signatures
        # PGP-based registry signatures were deprecated on April 25th, 2023
        try:
            response = requests.get(f"https://registry.npmjs.org/{package}", timeout=20)
            response.raise_for_status()

            data = response.json()
            version_info = data.get("versions", {}).get(package_version, {})

            # Check for signature in dist metadata
            dist_info = version_info.get("dist", {})
            signatures = dist_info.get("signatures", [])

            if signatures:
                valid_signatures = [sig for sig in signatures if sig.get("keyid") and sig.get("sig")]
                return {"signature_present": True, "signature_valid": len(valid_signatures) > 0}

            return {"signature_present": False, "signature_valid": False}

        except requests.RequestException as e:
            logging.error(f"Error checking NPM signature: {str(e)}")
            return {"signature_present": False, "signature_valid": False}

    if pm == "maven":
        return check_maven_signature(package_name, package_version)
    elif pm in ("yarn-berry", "yarn-classic", "pnpm", "npm"):
        return check_npm_signature(package_name, package_version)
    else:
        # log stuff
        # blow up
        logging.error(f"Package manager {pm} not supported.")


def api_constructor(package_name, repository):
    repo_url = repository.replace("https://", "").replace("http://", "").replace("/issues", "")

    # simplified_path = repo_url.replace("github.com/", "").split('#')[0].split('tree/master')[0].rstrip('/')
    repo_url = repo_url.split("#")[0]
    match = re.search(r"github\.com[:/](.*?)/(.*?)(?:/|$)", repo_url)
    if match:
        simplified_path = f"{match.group(1)}/{match.group(2)}"
    else:
        simplified_path = ""

    if simplified_path.endswith(".git"):
        simplified_path = simplified_path[:-4]

    repo_api = f"https://api.github.com/repos/{simplified_path}"

    error_message = None

    try:
        package_name = package_name.replace("npm:", "")
        parts = package_name.split("@")
        package_full_name = None
        name = None
        version = None

        if len(parts) > 2 or package_name.startswith("@"):
            package_full_name = f"@{parts[1]}"
            _, name = parts[1].split("/")  # scope,name
            version = parts[2]

        elif len(parts) == 2:
            if "/" in parts[1]:
                package_full_name = parts[0]
                scope, name = parts[0].split("/")
                version = parts[1]
            else:
                package_full_name = parts[0]
                name = parts[0]
                version = parts[1]

    except (ValueError, TypeError, AttributeError) as e:
        error_message = f"[INFO][api_constructor] Error: {str(e)}"
    except Exception as e:
        error_message = f"Unexpected error: {str(e)}"

    return repo_api, simplified_path, package_full_name, name, version, error_message


def check_parent_scm(package):
    name, version = package.split("@")
    group_id, artifact_id = name.split(":")

    existing_scm_data, repo_api, simplified_path, package_full_name = None, None, None, None
    stopping = False
    while not stopping:
        # First, getting the parent's pom contents
        command = [
            "mvn",
            "org.apache.maven.plugins:maven-help-plugin:3.5.1:evaluate",
            "-Dexpression=project.parent",
            f"-Dartifact={group_id}:{artifact_id}:{version}",
            "-q",
            "-DforceStdout",
        ]
        output = subprocess.run(command, capture_output=True, text=True)
        parent_pom = output.stdout.strip()
        if not parent_pom or "null" in parent_pom:
            # If there's no parent, we stop
            stopping = True
        else:
            parents_contents = xmltodict.parse(parent_pom)
            parent_group_id, parent_artifact_id = [
                parents_contents.get("project", {}).get("groupId", ""),
                parents_contents.get("project", {}).get("artifactId", ""),
            ]
            if not parent_group_id or not parent_artifact_id or parent_group_id != group_id:
                # If the parent is lacking data we stop;
                # If the parent doesn't share the same group, we went too far, so we stop too
                stopping = True
                break
            parent_scm_locations = [
                parents_contents.get("project", {}).get("scm", {}).get(location, "")
                for location in ["url", "connection", "developerConnection"]
            ] + [parents_contents.get("project", {}).get("url", "")]
            for location in parent_scm_locations:
                if location:
                    repo_api, simplified_path, package_full_name, _, _, _ = api_constructor(package, location)
                    data = make_github_request(repo_api, max_retries=2)
                    if data:
                        stopping = True
                        existing_scm_data = data
                        break
            if not stopping:
                group_id, artifact_id = parent_group_id, parent_artifact_id

    return {
        "data": existing_scm_data,
        "repo_api": repo_api,
        "simplified_path": simplified_path,
        "package_full_name": package_full_name,
    }


def check_source_code_by_version(package_name, version, repo_api, repo_link, simplified_path, package_manager):
    def check_git_head_presence(package_name, version):
        # In NPM-based packages, the registry may contain a gitHead field in the package's metadata
        # Although it's not mandatory to have it, if it's present, it's the best way to check
        # the package's source code for the specific version
        try:
            response = requests.get(f"https://registry.npmjs.org/{package_name}/{version}", timeout=20)
            response.raise_for_status()
            data = response.json()
            git_head = data.get("gitHead", "")
            return git_head
        except requests.RequestException as e:
            logging.error(f"Error checking gitHead presence: {str(e)}")
            return False

    source_code_info = {
        "exists": False,
        "tag_version": version,
        "is_sha": False,
        "sha": None,
        "url": None,
        "message": "No tags found in the repo",
        "status_code": 404,
    }
    if package_manager in ["yarn-berry", "yarn-classic", "pnpm", "npm"]:
        if git_head := check_git_head_presence(package_name, version):
            try:
                response = make_github_request(f"{repo_api}/commits/{git_head}", max_retries=2)
                if response:
                    logging.info(f"gitHead {git_head} found in {repo_link}")
                    return {
                        "exists": True,
                        "tag_version": version,
                        "is_sha": True,
                        "sha": git_head,
                        "url": None,
                        "message": "gitHead found in package metadata",
                        "status_code": 200,
                    }
                else:
                    logging.warning(f"gitHead {git_head} not found in {repo_link}, checking tags")
                    source_code_info = {
                        "exists": False,
                        "tag_version": version,
                        "is_sha": True,
                        "sha": git_head,
                        "url": None,
                        "message": f"gitHead {git_head} not found in {repo_link}",
                        "status_code": 404,
                    }
            except Exception as e:
                logging.error(f"Error checking gitHead in repo: {str(e)}")
        else:
            logging.warning(f"gitHead not found in {package_name} {version} metadata")
    else:
        logging.warning(
            f"Package manager {package_manager} not supported for gitHead checking, will proceed with tags"
        )

    have_no_tags_check_api = f"{repo_api}/tags"
    have_no_tags_response = requests.get(have_no_tags_check_api, headers=headers)
    have_no_tags_response_status_code = have_no_tags_response.status_code
    have_no_tags_data = have_no_tags_response.json()

    release_tag_exists = False
    if len(have_no_tags_data) == 0:
        logging.warning(f"No tags found for {package_name} in {repo_api}")
        release_tag_url = None
        message = "No tags found in the repo"
        status_code_release_tag = have_no_tags_response_status_code
    else:
        tag_possible_formats = construct_tag_format(version, package_name, repo_name=simplified_path)
        existing_tag_format = find_existing_tags_batch(tag_possible_formats, simplified_path)
        logging.info(f"Existing tag format: {existing_tag_format}")
        if existing_tag_format:
            existing_tag_format = existing_tag_format[0]
            release_tag_exists = True
            release_tag_url = f"{repo_api}/git/ref/tags/{existing_tag_format}"
            message = f"Tag {existing_tag_format} is found in the repo"
            status_code_release_tag = 200
        else:
            logging.warning(f"Tag {version} not found in {repo_api}")
            release_tag_url = None
            message = f"Tag {version} not found in the repo"
            status_code_release_tag = 404

    source_code_info.update(
        {
            "exists": release_tag_exists,
            "tag_version": version,
            "url": release_tag_url,
            "message": message,
            "status_code": status_code_release_tag,
        }
    )

    return source_code_info


def check_existence(package_name, repository, extract_message, package_manager):
    """Check if the package exists in the repository."""
    if "Could not find repository" in extract_message:
        return {"is_github": False, "github_url": "No_repo_info_found"}
    elif "Not a GitHub repository" in extract_message:
        return {"is_github": False, "github_url": repository}

    repo_api, simplified_path, package_full_name, _, version, error_message = api_constructor(package_name, repository)

    repo_link = f"https://github.com/{simplified_path}".lower()
    github_exists = False
    archived = False
    is_fork = False
    release_tag_exists = False
    release_tag_url = None
    message = None
    status_code_release_tag = None
    github_redirected = False
    now_repo_url = None
    open_issues_count = None
    status_code = 404

    data = make_github_request(repo_api, max_retries=2)
    parent_scm_result = {}
    source_code_info = None
    if not data:
        if package_manager == "maven":
            # There's the possibility of, in maven's case, assembly inheritance not having worked well;
            # As such, if the package manager is maven, we'll try to "work our way up", and perform the same check in the parent
            parent_scm_result = check_parent_scm(package_name)

    if not data and not parent_scm_result.get("data"):
        # simplified_path = parent_scm_result.get("simplified_path", simplified_path)
        # If we went up, and there's no still data, there really isn't a findable repository
        logging.warning(f"No repo found for {package_name} in {repo_link}")
        archived = None
        is_fork = None
        repo_link = f"https://github.com/{simplified_path}".lower()
        source_code_info = {
            "exists": release_tag_exists,
            "tag_version": version,
            "url": release_tag_url,
            "message": message,
            "status_code": status_code_release_tag,
        }
    else:
        data = data or parent_scm_result["data"]
        simplified_path = parent_scm_result.get("simplified_path", simplified_path)
        repo_api = parent_scm_result.get("repo_api", repo_api)
        package_full_name = parent_scm_result.get("package_full_name", package_full_name)

        status_code = 200
        github_exists = True
        open_issues_count = data["open_issues"]
        if data["archived"]:
            archived = True
        if data["fork"]:
            is_fork = True
            # Get the parent repo link
            parent_repo_link = data.get("parent", {}).get("html_url", "")

        now_repo_url = data.get("html_url", "").lower()
        repo_link = f"https://github.com/{simplified_path}".lower()

        if now_repo_url != repo_link:
            github_redirected = True
        else:
            now_repo_url = None

        source_code_info = check_source_code_by_version(
            package_full_name, version, repo_api, repo_link, simplified_path, package_manager
        )

    github_info = {
        "is_github": True,
        "github_api": repo_api,
        "github_url": repo_link,
        "github_exists": github_exists,
        "github_redirected": github_redirected,
        "redirected_repo": now_repo_url,
        "status_code": status_code,
        "archived": archived,
        "is_fork": is_fork,
        "source_code_version": source_code_info,
        "parent_repo_link": parent_repo_link if is_fork else None,
        "open_issues_count": open_issues_count,
        "error": error_message if error_message else "No error message.",
    }

    return github_info


def get_api_content(api, headers):
    """
    Get the content of the API.
    """
    try:
        response = requests.get(api, headers=headers)
        response.raise_for_status()
        content = response.json()
        file_content = base64.b64decode(content["content"]).decode("utf-8")
        package_json = json.loads(file_content)
        return package_json

    except (
        requests.HTTPError,
        requests.ConnectionError,
        requests.Timeout,
        json.JSONDecodeError,
    ) as e:
        logging.error(f"Request error: {str(e)} for URL: {api}")
        return None
    except Exception as e:
        logging.error(f"Unexpected error: {str(e)} for URL: {api}")
        return None


def check_name_match_for_fork(package_name, repository):
    name_match = False
    unmatch_info = None

    repo_api, repo, _, pkg_only_name, _, _ = api_constructor(package_name, repository)
    fork_api_json_url = [
        f"{repo_api}/contents/package.json",
        f"{repo_api}/contents/contents/packages/{pkg_only_name}/package.json",
    ]

    fork_api_json = None
    api = ""
    status_code = None

    for api in fork_api_json_url:
        fork_api_response = requests.get(api, headers=headers, timeout=20)
        status_code = fork_api_response.status_code
        if status_code == 200:
            fork_api_json = fork_api_response.json()
            break

    if not fork_api_json:
        return {
            "repository": repository,
            "find_json_package": "",
            "repo_name": repo,
            "package_api_in_packages": api,
            "match": False,
            "unmatch_info": {
                "status_code": status_code,
                "reason": "API requests failed",
                "api_url": api,
            },
        }

    file_content = base64.b64decode(fork_api_json["content"]).decode("utf-8")
    package_json_file = json.loads(file_content)

    package_name_in_fork = package_json_file.get("name", "")
    if package_name_in_fork == package_name.rsplit("@", 1)[0]:
        name_match = True

    if not name_match:
        unmatch_info = {
            "status_code": status_code,
            "repo_name": repository,
            # 'total_count': search_results['total_count'],
            "package_name": package_name,
            "find_json_package": package_name_in_fork,
            "reason": "could not find",
            "api_url": api,
        }

    match_info = {
        "repository": repository,
        "find_json_package": package_name_in_fork,
        "package_name": package_name.rsplit("@", 1)[0],
        "repo_name": repo,
        "package_api_in_packages": api,
        "match": name_match,
        "unmatch_info": None if name_match else unmatch_info,
    }

    return match_info


def check_name_match(package_name, repository):
    cache_manager._setup_requests_cache(cache_name="static_analysis")

    _, repo_name, _, _, _, _ = api_constructor(package_name, repository)
    original_package_name = package_name.rsplit("@", 1)[0]

    query = f'"\\"name\\": \\"{original_package_name}\\""'
    repo = f"{repo_name}"

    encoded_search_term = urllib.parse.quote(query)
    encoded_repo = urllib.parse.quote(repo)

    url = f"https://api.github.com/search/code?q={encoded_search_term}+in:file+repo:{encoded_repo}"

    response = requests.get(url, headers=headers, timeout=20)

    # if not response.from_cache:
    #     time.sleep(6)

    status_code = response.status_code

    is_match = False
    unmatch_info = None
    package_api_in_packages = None

    if status_code != 200:
        search_results = None

    if status_code == 200:
        search_results = response.json()
        if search_results["total_count"] == 0:
            package_name = "can not find"

        else:
            for item in search_results["items"]:
                if item["name"] == "package.json":
                    package_api_in_packages = item["url"]
                    is_match = True

    if not is_match:
        unmatch_info = {
            "status_code": status_code,
            "repo_name": repository,
            # 'total_count': search_results['total_count'],
            "package_name": package_name,
            "search_term": original_package_name,
            "query": query,
            "reason": "could not find",
            "api_url": url,
        }

    match_info = {
        "repository": repository,
        "find_json_package": package_api_in_packages,
        "repo_name": repo,
        "package_api_in_packages": package_api_in_packages,
        "match": is_match,
        "unmatch_info": None if is_match else unmatch_info,
    }

    return match_info


def analyze_package_data(
    package, repo_url, extract_message, pm, check_match=False, enabled_checks=DEFAULT_ENABLED_CHECKS
):
    """
    Analyze package data with configurable smell checks.

    Args:
        package: Package to analyze
        repo_url: Repository URL
        extract_message: Message from repository URL extraction - is it or not a GitHub repository
        pm: Package manager
        check_match: Whether to check name matches
        enabled_checks: Dictionary of enabled smell checks
    """

    def cached_analysis_matches_schema(cached_analysis, schema):
        for key, value in schema.items():
            if isinstance(value, dict):
                if key not in cached_analysis:
                    return False
                if not cached_analysis_matches_schema(cached_analysis[key], value):
                    return False
            elif key not in cached_analysis:
                return False
        return True

    package_info = {}
    try:
        package_name, package_version = package.rsplit("@", 1)
        package_version = package_version.replace("npm:", "")

        # Try to get from cache first
        cached_analysis = cache_manager.package_cache.get_package_analysis(package_name, package_version, pm)

        # Initialize missing_checks to track what needs to be analyzed
        missing_checks = {}

        if cached_analysis:
            logging.info(f"Found cached analysis for {package}")
            package_info = cached_analysis

            # Check which enabled checks are missing from cache
            for check, enabled in enabled_checks.items():
                if enabled:
                    if check in ["source_code_sha", "forks"]:
                        check = "source_code"
                    elif check in ["deprecated", "provenance"]:
                        check = "package_info"
                    elif check == "aliased_packages":
                        continue
                    missing_checks[check] = not cached_analysis_matches_schema(
                        cached_analysis.get(check, {}), SCHEMAS_FOR_CACHE_ANALYSIS[check]
                    )

            if all(not missing for missing in missing_checks.values()):
                logging.info(f"Using complete cached analysis for {package}")
                return package_info
            logging.info(
                f"Found partial cached analysis for {package}, analyzing missing checks: {list(check for check, missing in missing_checks.items() if missing)}"
            )
        else:
            logging.info(f"No cached analysis for {package}, analyzing all enabled checks")
            for check, enabled in enabled_checks.items():
                if check in ["source_code_sha", "forks", "aliased_packages"]:
                    continue
                elif check in ["deprecated", "provenance"]:
                    check = "package_info"
                missing_checks[check] = enabled

        for check in missing_checks:
            if not missing_checks[check]:
                continue
            package_info[check] = SCHEMAS_FOR_CACHE_ANALYSIS[check].copy()

        if missing_checks.get("package_info"):
            package_infos = check_deprecated_and_provenance(package_name, package_version, pm)
            update_package_info(package_info, "package_info", package_infos)

        if missing_checks.get("code_signature"):
            update_package_info(
                package_info, "code_signature", check_code_signature(package_name, package_version, pm)
            )

        if missing_checks.get("source_code"):
            update_package_info(package_info, "source_code", check_existence(package, repo_url, extract_message, pm))

        if check_match and package_info.get("source_code") and package_info["source_code"].get("github_exists"):
            repo_url_to_use = package_info["source_code"].get("redirected_repo") or repo_url
            if package_info.get("provenance") == False:
                if (
                    package_info["source_code"].get("is_fork") == True
                    or package_info["source_code"].get("archived") == True
                ):
                    package_info["match_info"] = check_name_match_for_fork(package, repo_url_to_use)
                else:
                    package_info["match_info"] = check_name_match(package, repo_url_to_use)
            else:
                package_info["match_info"] = {
                    "has_provenance": True,
                    "match": True,
                    "repo_name": repo_url,
                }

        # Cache the updated analysis
        cache_manager.package_cache.cache_package_analysis(package_name, package_version, pm, package_info)

    except Exception as e:
        logging.error(f"Analyzing package {package}: {str(e)}")
        package_info["error"] = str(e)

    return package_info


def disable_checks_from_config(package_name, parent, config, enabled_checks):
    """
    Returns the enabled_checks dictionary for the package, based on the configuration file.
    config[<key>] includes a series of entries (regex patterns) which specify which packages to ignore/do less checks on.
    We compare the package name against these patterns.
    If there are conflicting patterns, the first one that matches is used.

    Args:
        package_name (str): Name of the package
        parent(str): Name of the package's parent
        config (dict): Configuration dictionary
        enabled_checks (dict): Dictionary of enabled checks

    Returns:
        dict: Package-specific enabled checks
    """
    possible_keys = [[package_name, "ignore"], [parent, "ignore-if-parent"]]
    possible_keys = [info for info in possible_keys if info[1] in config]
    if not config or not possible_keys:
        logging.warning(f"possible_keys: {possible_keys}")
        logging.warning("No config file provided, using default config (no packages ignored)")
        return enabled_checks

    for name, config_type in possible_keys:
        if not name:
            continue
        for pattern in config[config_type]:
            try:
                if re.match(pattern, name):
                    if isinstance(config[config_type][pattern], str):
                        if config[config_type][pattern] == "all":
                            logging.info(f"Ignoring all checks for {package_name}")
                            return {}
                    elif isinstance(config[config_type][pattern], list):
                        for check in config[config_type][pattern]:
                            logging.info(f"Ignoring check {check} for {package_name}")
                            enabled_checks[check] = False
                    else:
                        logging.warning(f"Invalid ignore pattern for {package_name}: {config[config_type][pattern]}")
                    break
            except Exception as e:
                logging.error(f"Error parsing config file patterns: {e}; pattern: {pattern}, name: {name}")
    logging.info(f"The following enabled_checks will be set for this package: {enabled_checks}")
    return enabled_checks


def get_static_data(folder, packages_data, pm, check_match=False, enabled_checks=DEFAULT_ENABLED_CHECKS, config=None):
    logging.info("Analyzing package static data...")
    package_all = {}
    errors = {}
    with tqdm(total=len(packages_data), desc="Analyzing packages") as pbar:
        for package, repo_urls in packages_data.items():
            logging.info(f"Currently analyzing {package}")
            repo_url = repo_urls.get("url", "")
            extract_repo_url_message = repo_urls.get("message", "")
            command = repo_urls.get("command", None)
            parent = repo_urls.get("parent", "")

            package_enabled_checks = disable_checks_from_config(package, parent, config, enabled_checks)
            if not package_enabled_checks:
                logging.warning(f"Package {package} will be skipped, no checks enabled for it")
                pbar.update(1)
                continue

            analyzed_data = analyze_package_data(
                package, repo_url, extract_repo_url_message, pm, check_match=check_match, enabled_checks=enabled_checks
            )
            error = analyzed_data.get("error", None)
            pbar.update(1)

            if error:
                errors[package] = error
            else:
                package_all[package] = analyzed_data
                package_all[package]["parent"] = parent
                package_all[package]["command"] = repo_urls.get("command", None)

    return package_all, errors


def save_results_to_file(filename, data):
    with open(filename, "w", encoding="utf-8") as file:
        json.dump(data, file, indent=4)



import os
import subprocess
import re
import json
import sqlite3
import logging
import requests
import xmltodict
from pathlib import Path
from tqdm import tqdm
from tool.tool_config import get_cache_manager
from typing import List

TIMEOUT = 60

cache_manager = get_cache_manager()
GITHUB_URL_PATTERN = re.compile(r"(github.*)", re.IGNORECASE)


def write_output(folder_path, filename, data):
    folder_path = os.path.join(folder_path, "dep_lists")
    os.makedirs(folder_path, exist_ok=True)

    file_path = os.path.join(folder_path, filename)
    # check if the folder exists

    with open(file_path, "w", encoding="utf-8") as f:
        if isinstance(data, list):
            f.write("\n".join(data))
            f.write(f"\nTotal: {len(data)}\n")
        else:
            json.dump(data, f, indent=2)


def extract_repo_url(repo_info: str) -> str:
    """Extract GitHub repository URL from repository information."""
    url = repo_info
    if "https" not in repo_info:
        # cases such as git@github:apache/maven-scm, we just remove the :
        url = url.replace(":/", "/")
    else:
        # could be a redirect, so we'll make a request to the URL to get the final URL
        match = GITHUB_URL_PATTERN.search(url)
        if not match:
            try:
                url = requests.head(url, allow_redirects=True).url
            except requests.exceptions.RequestException:
                logging.warning(f"Could not check for redirections, was using {url}")
    url = url.replace(":", "/")
    match = GITHUB_URL_PATTERN.search(url)
    if not match:
        return repo_info, "Not a GitHub repository"

    # if there is a match, there's still the possibility of the scm url having been
    # put in a different form, e.g.,
    # github.com/apache/maven-scm/tree/maven-scm-2.1.0/maven-scm-providers/maven-scm-providers-standard
    # from here, we only want the URL up until the second-most directory after github.com
    url = match.group(0)
    parts = url.split("/")
    joined = "/".join(parts[:3]) if len(parts) > 3 else url
    joined = joined if not joined.endswith(".git") else joined[:-4]
    return joined, "GitHub repository"


def get_scm_command(pm: str, package: str) -> List[str]:
    """Get the appropriate command to find a package's source code locations for the package manager."""
    if pm == "yarn-berry" or pm == "yarn-classic":
        return ["yarn", "info", package.replace("@npm:", "@"), "repository.url", "--silent"]
    elif pm == "pnpm":
        # for cases like @babel/helper-create-class-features-plugin@7.25.9(@babel/core@7.26.10),
        # we look for the repository of the package inside parentheses
        if "(" in package:
            package = package.split("(")[1].split(")")[0]
        return ["pnpm", "info", package.replace("@npm:", "@"), "repository.url"]
    elif pm == "npm":
        return ["npm", "info", package.replace("@npm:", "@"), "repository.url"]
    elif pm == "maven":
        name, version = package.split("@")
        group_id, artifact_id = name.split(":")
        return [
            "mvn",
            "org.apache.maven.plugins:maven-help-plugin:3.5.1:evaluate",
            f"-Dexpression=project",
            f"-Dartifact={group_id}:{artifact_id}:{version}",
            "-q",
            "-DforceStdout",
        ]
    raise ValueError(f"Unsupported package manager: {pm}")


def run_scm_command(pm, command):
    def run_npm_command(command):
        try:
            result = subprocess.run(
                command,
                capture_output=True,
                text=True,
                check=True,
                timeout=TIMEOUT,
            )
            return result.stdout
        except subprocess.TimeoutExpired:
            logging.warning(
                f"Command {command} timed out after {TIMEOUT} seconds",
            )
            return None
        except subprocess.CalledProcessError as e:
            logging.warning(f"Command {command} failed: {e}")
            return None

    def run_maven_command(command):
        try:
            result = subprocess.run(
                command,
                capture_output=True,
                text=True,
                check=True,
                timeout=TIMEOUT,
            )
            output = result.stdout
            if output:
                parsed_content = xmltodict.parse(output)
                locations = [
                    parsed_content.get("project", {}).get("scm", {}).get("url", ""),
                    parsed_content.get("project", {}).get("scm", {}).get("connection", ""),
                    parsed_content.get("project", {}).get("scm", {}).get("developerConnection", ""),
                    parsed_content.get("project", {}).get("url", ""),
                ]
                return next((loc for loc in locations if loc), None)

        except subprocess.TimeoutExpired:
            logging.warning(
                f"Command {command} timed out after {TIMEOUT} seconds",
            )
            return None
        except subprocess.CalledProcessError as e:
            logging.warning(f"Command {command} failed: {e}")
            return None

    if pm == "npm" or pm == "yarn-berry" or pm == "yarn-classic" or pm == "pnpm":
        return run_npm_command(command)
    elif pm == "maven":
        return run_maven_command(command)
    raise ValueError(f"Unsupported package manager: {pm}")


def process_package(
    package,
    parent,
    command,
    pm,
    repos_output,
    undefined,
    same_repos_deps,
    some_errors,
    repos_output_json,
):
    def check_if_valid_repo_info(repo_info):
        retrieved_info = {
            "url": "",
            "parent": "",
            "message": "",
            "command": "",
        }
        if (
            repo_info is None
            or "Undefined" in repo_info
            or "undefined" in repo_info
            or "ERR!" in repo_info
            or "null object" in repo_info
        ):
            logging.warning(f"Could not find repository for {package}")
            retrieved_info.update(
                {
                    "url": "Could not find",
                    "parent": parent,
                    "message": "Could not find repository",
                    "command": command,
                }
            )
            repos_output_json[package] = retrieved_info
            undefined.append(f"Undefined for {package}, {repo_info}")
            return False, retrieved_info

        url, message = extract_repo_url(repo_info)
        retrieved_info.update(
            {
                "url": url,
                "parent": parent,
                "message": message,
                "command": command,
            }
        )
        repos_output_json[package] = retrieved_info
        if message == "GitHub repository":
            logging.info(f"Found GitHub URL for {package}: {url}")
            repos_output.append(url)
            same_repos_deps.get("url", []).append(package)
            return True, retrieved_info
        else:
            logging.info(f"Found non-GitHub URL for {package}: {url}")
            some_errors.append(f"No GitHub URL for {package}\n{repo_info}")
            return False, retrieved_info

    retrieved_info = cache_manager.github_cache.get_github_url(package)
    if not retrieved_info:
        result = run_scm_command(pm, get_scm_command(pm, package))
        if result:
            valid_repo_info, retrieved_info = check_if_valid_repo_info(result)
        else:
            logging.warning(f"SCM command failed for {package}")
            retrieved_info = {
                "url": "Could not find",
                "parent": parent,
                "message": "Could not find repository",
                "command": command,
            }
            repos_output_json[package] = retrieved_info
            undefined.append(f"Undefined for {package}, {result}")

        cache_manager.github_cache.cache_github_url(package, retrieved_info)
    else:
        logging.info(f"Found cached URL for {package}: {retrieved_info.get('url', '-')}")
        valid_repo_info = "GitHub repository" == retrieved_info["message"]
        if parent != retrieved_info.get("parent", None):
            # tackles scenarios where parent was cached and then removed; incoming parent should take precedence
            retrieved_info.update({"parent": parent})
            cache_manager.github_cache.cache_github_url(package, retrieved_info)
        repos_output_json[package] = retrieved_info
        if valid_repo_info:
            repos_output.append(retrieved_info["url"])
            same_repos_deps.get("url", []).append(package)
        else:
            some_errors.append(f"No GitHub URL for {package}\n{retrieved_info['url']}")


def get_github_repo_url(folder, dep_list, pm):
    repos_output = []  # List to store GitHub URLs
    some_errors = []  # List to store packages without a GitHub URL
    undefined = []  # List to store packages with undefined repository URLs
    same_repos_deps = {}  # Dict to store packages with same GitHub URL
    repos_output_json = {}  # Dict to store packages with GitHub URL

    logging.info("Getting GitHub URLs of packages...")
    total_packages_to_process = len(dep_list.get("resolutions", []))
    # have not process patches
    with tqdm(total=total_packages_to_process, desc="Getting GitHub URLs") as pbar:
        for pkg_res in dep_list.get("resolutions"):
            package = pkg_res["info"]
            parent = pkg_res.get("parent", None)
            command = pkg_res.get("command", None)
            process_package(
                package,
                parent,
                command,
                pm,
                repos_output,
                undefined,
                same_repos_deps,
                some_errors,
                repos_output_json,
            )
            pbar.update(1)

    # Write collected data to files
    unique_repos_output = sorted(set(repos_output))
    outputs = {
        # "github_repo_all.txt": repos_output,
        # "github_repo_unique.txt": unique_repos_output,
        "github_repo_undefined.log": undefined,
        # "github_repo_some_error.log": some_errors,
        "github_repos_depsnsamerepo.json": same_repos_deps,
        "github_repos_output.json": repos_output_json,
    }

    for filename, data in outputs.items():
        write_output(folder, filename, data)

    return repos_output_json



import re
import logging

MESSAGE_TO_VERSIONS_MAPPING = {
    "newly_added": "Newly added package",
    "deleted": "Deleted package",
    "upgraded": "Upgraded package",
    "downgraded": "Downgraded package",
    "no_change": "No change",
}


def parse_dependencies(file_path):
    dependencies = {}
    with open(file_path, "r") as file:
        for line in file:
            if "@" in line:
                dep, version = line.strip().rsplit("@", 1)
                if dep in dependencies:
                    dependencies[dep].append(version)
                else:
                    dependencies[dep] = [version]

    return dependencies


def sort_versions(versions):
    def parse_version(v):
        try:
            major, minor, patch = map(int, v.split("."))
            return major, minor, patch
        except ValueError:
            return float("-inf"), v

    return sorted(versions, key=parse_version)


def compare_dependencies(deps1, deps2):
    logging.info("Comparing dependencies...")
    differences_pkg = {}
    all_deps = set(list(deps1.keys()) + list(deps2.keys()))

    # 1 is old, 2 is new
    for dep in all_deps:
        versions1 = deps1.get(dep, [])
        versions2 = deps2.get(dep, [])

        if versions1 and versions2:
            only_in_1 = set(versions1) - set(versions2)
            only_in_2 = set(versions2) - set(versions1)
            unchanged_in_1 = set(versions1) - only_in_1

            if only_in_1 or only_in_2:
                differences_pkg[dep] = {
                    "only_in_1": list(sort_versions(only_in_1)),
                    "only_in_2": list(sort_versions(only_in_2)),
                    "unchanged": list(sort_versions(unchanged_in_1)),
                    "version1": list(sort_versions(versions1)),
                    "version2": list(sort_versions(versions2)),
                }

        else:
            differences_pkg[dep] = {
                "version1": list(sort_versions(versions1)),
                "version2": list(sort_versions(versions2)),
            }

    return differences_pkg


def choose_compare_version(dep_file_1, dep_file_2):
    differences = compare_dependencies(dep_file_1, dep_file_2)

    for dep, versions in differences.items():
        differences[dep]["message"] = None
        if not versions.get("only_in_2"):
            chosen_v2 = None
            if versions.get("only_in_1") is not None:
                differences[dep]["message"] = f"Deleted package - version {versions.get('only_in_1')} deleted"

        else:
            chosen_v2 = versions["only_in_2"][-1]

        if not versions.get("only_in_1"):
            unchanged = versions.get("unchanged")
            if not unchanged:
                chosen_v1 = None
            else:
                chosen_v1 = versions.get("unchanged")[-1]
        else:
            chosen_v1 = versions["only_in_1"][-1]

        differences[dep]["chosen_v2"] = chosen_v2
        differences[dep]["chosen_v1"] = chosen_v1

    return differences


def is_version_greater(v1, v2):
    v1_nums = [int(num) for num in v1.split(".") if num.isdigit()]
    v2_nums = [int(num) for num in v2.split(".") if num.isdigit()]

    return v1_nums > v2_nums


def category_dependencies(dep_file_1, dep_file_2):
    def categorize_dependency(v1, v2, allv1, allv2):
        if v1 is None or v2 is None:
            if not allv1:
                return "newly_added", allv1, allv2
            elif not allv2:
                return "deleted", allv1, allv2
        else:
            if is_version_greater(v2, v1):
                return "upgraded", v1, v2
            elif is_version_greater(v1, v2):
                return "downgraded", v1, v2
            else:
                return "no_change", v1, v2
        return None

    differences = choose_compare_version(dep_file_1, dep_file_2)
    gathered_categories = {
        "newly_added": {},
        "deleted": {},
        "upgraded": {},
        "downgraded": {},
        "no_change": {},
    }

    for dep, versions in differences.items():
        v1, v2 = versions["chosen_v1"], versions["chosen_v2"]
        allv1, allv2 = versions["version1"], versions["version2"]

        categorized_dependency = categorize_dependency(v1, v2, allv1, allv2)
        if categorized_dependency:
            category, v1, v2 = categorized_dependency
            message = MESSAGE_TO_VERSIONS_MAPPING[category]
            differences[dep]["message"] = message
            gathered_categories[category][dep] = {
                "version1": v1,
                "version2": v2,
                "message": message,
            }

    return (
        differences,
        *gathered_categories.values(),
    )


def get_repo_from_SA(dep_file_1, dep_file_2, SA_old, SA_new):
    (
        differences,
        newly_added_pkg,
        deleted_pkg,
        upgraded_pkg,
        downgraded_pkg,
        no_change_pkg,
    ) = category_dependencies(dep_file_1, dep_file_2)

    for dep, versions in differences.items():
        differences[dep]["v1_repo_accessibility"] = None
        differences[dep]["v2_repo_accessibility"] = None
        differences[dep]["v1_repo_link"] = None
        differences[dep]["v2_repo_link"] = None
        differences[dep]["repo_message"] = None
        differences[dep]["compare_message"] = None
        differences[dep]["repo"] = None
        differences[dep]["repo_name"] = None
        differences[dep]["repo_pure"] = None
        differences[dep]["v1_repo_directed"] = None
        differences[dep]["v2_repo_directed"] = None

        if versions["message"] == "Upgraded package":
            differences[dep]["category"] = "Upgraded package"
            # Check code signature changes
            signature_changes = compare_code_signatures(
                dep, versions["chosen_v1"], versions["chosen_v2"], SA_old, SA_new
            )
            differences[dep]["signature_changes"] = signature_changes

            # If there are signature changes, add it to the category
            if signature_changes["has_changes"]:
                differences[dep]["category"] = "Upgraded package with signature changes"

            dep_name_version_old = f"{dep}@{versions['chosen_v1']}"
            dep_name_version_new = f"{dep}@{versions['chosen_v2']}"

            dep_old_repo_url_accessibility = (
                SA_old.get(dep_name_version_old, {}).get("source_code", {}).get("github_exists", None)
            )
            dep_new_repo_url_accessibility = (
                SA_new.get(dep_name_version_new, {}).get("source_code", {}).get("github_exists", None)
            )

            if dep_old_repo_url_accessibility is True and dep_new_repo_url_accessibility is True:
                dep_old_repo_url = (
                    SA_old.get(dep_name_version_old, {}).get("source_code", {}).get("github_url", "Error")
                )
                dep_new_repo_url = (
                    SA_new.get(dep_name_version_new, {}).get("source_code", {}).get("github_url", "Error")
                )

                differences[dep]["v1_repo_accessibility"] = dep_old_repo_url_accessibility
                differences[dep]["v2_repo_accessibility"] = dep_new_repo_url_accessibility

                differences[dep]["v1_repo_link"] = dep_old_repo_url
                differences[dep]["v2_repo_link"] = dep_new_repo_url

                if dep_old_repo_url != dep_new_repo_url:
                    differences[dep]["repo_message"] = "Different repo link"
                    differences[dep]["compare_message"] = "DO NOT COMPARE"

                    check_directed_old = (
                        SA_old.get(dep_name_version_old, {}).get("source_code", {}).get("github_redirected", None)
                    )
                    check_directed_new = (
                        SA_new.get(dep_name_version_new, {}).get("source_code", {}).get("github_redirected", None)
                    )

                    differences[dep]["v1_repo_directed"] = check_directed_old
                    differences[dep]["v2_repo_directed"] = check_directed_new

                else:
                    pattern = r"(github.*)"
                    match1 = re.search(pattern, dep_old_repo_url, re.IGNORECASE)

                    v1_repo_url_pure = match1.group(1) if match1 else "not github"
                    v1_repo_url_clean = (
                        v1_repo_url_pure.replace("https://", "").replace("http://", "").replace("/issues", "")
                    )

                    match2 = re.search(pattern, dep_new_repo_url, re.IGNORECASE)

                    v2_repo_url_pure = match2.group(1) if match2 else "not github"
                    v2_repo_url_clean = (
                        v2_repo_url_pure.replace("https://", "").replace("http://", "").replace("/issues", "")
                    )

                    if v1_repo_url_clean == v2_repo_url_clean:
                        repo_name = (
                            v1_repo_url_clean.replace("github.com/", "")
                            .split("#")[0]
                            .split("tree/master")[0]
                            .rstrip("/")
                        )
                        if repo_name.endswith(".git"):
                            repo_name = repo_name[:-4]
                        differences[dep]["repo"] = dep_old_repo_url
                        differences[dep]["repo_name"] = repo_name
                        differences[dep]["repo_pure"] = v1_repo_url_pure
                        differences[dep]["compare_message"] = "COMPARE"
                        differences[dep]["repo_message"] = "Repo - ok"

                    else:
                        differences[dep]["repo_message"] = "Different repo name"
                        differences[dep]["compare_message"] = "DO NOT COMPARE"

            else:
                differences[dep]["compare_message"] = "DO NOT COMPARE"
                if dep_old_repo_url_accessibility is False or dep_old_repo_url_accessibility is None:
                    differences[dep]["v1_repo_accessibility"] = "Repo link not accessible"
                    differences[dep]["compare_message"] = "DO NOT COMPARE"

                if dep_new_repo_url_accessibility is False or dep_new_repo_url_accessibility is None:
                    differences[dep]["v2_repo_accessibility"] = "Repo link not accessible"
                    differences[dep]["compare_message"] = "DO NOT COMPARE"
        else:
            differences[dep]["compare_message"] = "DO NOT COMPARE"
            dep_name_version_old = f"{dep}@{versions['chosen_v1']}"
            dep_name_version_new = f"{dep}@{versions['chosen_v2']}"
            repo_1 = SA_old.get(dep_name_version_old, {}).get("source_code", {}).get("github_url", "Error")
            repo_2 = SA_old.get(dep_name_version_old, {}).get("source_code", {}).get("github_url", "Error")
            if repo_1 == repo_2 and repo_1 != "Error" and repo_2 != "Error":
                differences[dep]["repo_message"] = "Repo - ok"
                differences[dep]["repo"] = repo_1
            elif repo_1 != "Error":
                differences[dep]["repo"] = repo_1
            elif repo_2 != "Error":
                differences[dep]["repo"] = repo_2

            differences[dep]["category"] = versions["message"]

            if versions["message"] == "Downgraded package":
                # Check code signature changes
                signature_changes = compare_code_signatures(
                    dep, versions["chosen_v1"], versions["chosen_v2"], SA_old, SA_new
                )
                differences[dep]["signature_changes"] = signature_changes

                # If there are signature changes, add it to the category
                if signature_changes["has_changes"]:
                    differences[dep]["category"] = "Downgraded package with signature changes"

    return (
        differences,
        newly_added_pkg,
        deleted_pkg,
        upgraded_pkg,
        downgraded_pkg,
        no_change_pkg,
    )


def changed_patch(package_data_old, package_data_new):
    logging.info("Comparing patches...")
    patches_change = {}
    no_change_patches = {}
    if package_data_old and package_data_new:
        common_patches = set(package_data_old.keys()) & set(package_data_new.keys())
        unique_patches_in_data2 = set(package_data_new.keys()) - set(package_data_old.keys())

        for new_patch in unique_patches_in_data2:
            patches_change[new_patch] = {
                "name": new_patch,
                "version": package_data_new[new_patch].get("version"),
                "patch_file_path": package_data_new[new_patch].get("patch_file_path"),
            }

        no_change_patches["No change patches"] = list(common_patches)
    else:
        patches_change["No packages patch found"] = {
            "name": None,
            "version": None,
            "patch_file_path": None,
        }
        no_change_patches["No change patches"] = []

    return patches_change, no_change_patches


def differential(dep_file_1, dep_file_2, SA_1, SA_2):
    compare_differences = choose_compare_version(dep_file_1, dep_file_2)

    (
        differences_pkg_full,
        newly_added_pkg,
        deleted_pkg,
        upgraded_pkg,
        downgraded_pkg,
        no_change_pkg,
    ) = get_repo_from_SA(dep_file_1, dep_file_2, SA_1, SA_2)

    return (
        compare_differences,
        differences_pkg_full,
        downgraded_pkg,
        newly_added_pkg,
        deleted_pkg,
        upgraded_pkg,
        no_change_pkg,
    )


def compare_code_signatures(pkg_name, old_version, new_version, SA_1, SA_2):
    """Compare code signatures between versions of a package."""
    old_pkg = f"{pkg_name}@{old_version}"
    new_pkg = f"{pkg_name}@{new_version}"

    old_signature = SA_1.get(old_pkg, {}).get("code_signature", {})
    new_signature = SA_2.get(new_pkg, {}).get("code_signature", {})

    changes = {
        "old_signature_present": old_signature.get("signature_present", False),
        "new_signature_present": new_signature.get("signature_present", False),
        "old_signature_valid": old_signature.get("signature_valid", False),
        "new_signature_valid": new_signature.get("signature_valid", False),
        "has_changes": False,
    }

    # Check if there are any changes in signature status
    if (
        changes["old_signature_present"]
        and not changes["new_signature_present"]
        or changes["old_signature_valid"]
        and not changes["new_signature_valid"]
    ):
        changes["has_changes"] = True

    return changes
